﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace PuzzleSolve
{
    public enum PuzzleSideLocation
    {
        Top = 0,
        Right = 1,
        Bottom = 2,
        Left = 3
    }

    public class PuzzleSolutionSet
    {
        public List<PuzzlePlacement[,]> Solutions = new List<PuzzlePlacement[,]>();
        public bool PreviouslyAttempted(PuzzlePlacement[,] thisAttempt)
        {
            foreach (PuzzlePlacement[,] PreviousAttempt in Solutions)
            {
                if (PuzzlePlacement.IsSame(PreviousAttempt, thisAttempt)) return true;
            }
            return false;
        }
    }

    public class PuzzlePiece
    {
        public PuzzleSide[] Side = new PuzzleSide[4];
        public int PieceId;

        public PuzzlePiece(int id, PuzzleSide a, PuzzleSide b, PuzzleSide c, PuzzleSide d)
        {
            PieceId = id;
            Side[0] = a; Side[1] = b; Side[2] = c; Side[3] = d;
        }

        public bool Match(int orientation, PuzzlePlacement Top, 
                         PuzzlePlacement Left, PuzzlePlacement Right,
                         PuzzlePlacement Bottom)
        {
            if (PuzzlePlacement.GetPuzzleSideFromPiece(this, orientation, PuzzleSideLocation.Top).MatchingSide(Top.GetPuzzleSide(PuzzleSideLocation.Bottom)) == true &&
                PuzzlePlacement.GetPuzzleSideFromPiece(this, orientation, PuzzleSideLocation.Left).MatchingSide(Left.GetPuzzleSide(PuzzleSideLocation.Right)) == true &&
                PuzzlePlacement.GetPuzzleSideFromPiece(this, orientation, PuzzleSideLocation.Right).MatchingSide(Right.GetPuzzleSide(PuzzleSideLocation.Left)) == true &&
                PuzzlePlacement.GetPuzzleSideFromPiece(this, orientation, PuzzleSideLocation.Bottom).MatchingSide(Bottom.GetPuzzleSide(PuzzleSideLocation.Top)))
            {
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    public class PuzzlePlacement
    {
        public PuzzlePiece Piece;
        public int Orientation;
        public PuzzlePlacement(PuzzlePiece piece, int orientation)
        {
            Piece = piece; Orientation = orientation;
        }

        public static PuzzleSide GetPuzzleSideFromPiece(PuzzlePiece piece, int orientation, int side)
        {
            return (piece.Side[((orientation+side)%4)]);
        }

        public static PuzzleSide GetPuzzleSideFromPiece(PuzzlePiece piece, int orientation, PuzzleSideLocation side)
        {
            return (GetPuzzleSideFromPiece(piece, orientation, (int)side));
        }

        public PuzzleSide GetPuzzleSide(int side)
        {
            return (GetPuzzleSideFromPiece(Piece, Orientation, side));
        }

        public PuzzleSide GetPuzzleSide(PuzzleSideLocation side)
        {
            return (GetPuzzleSide((int)side));
        }

        public static bool IsSame(PuzzlePlacement[,] a, PuzzlePlacement[,] b)
        {
            for (int x = 1; x < 4; x++)
            {
                for (int y = 1; y < 4; y++)
                {
                    if (a[x, y].Piece.PieceId != b[x, y].Piece.PieceId) return false;
                }
            }

            return true;
        }

        public static bool Used(PuzzlePlacement[,] PuzzleSolution, PuzzlePiece puzzlePiece)
        {
            for (int x = 1; x < 4; x++)
            {
                for (int y = 1; y < 4; y++)
                {
                    if (PuzzleSolution[x, y].Piece.PieceId == puzzlePiece.PieceId) return true;
                }
            }
            return false;
        }
    }

    public class PuzzleSide
    {
        public int Match;
        public bool Left;

        public PuzzleSide(int match, bool left)
        {
            Match = match; Left = left;
        }

        public bool MatchingSide(PuzzleSide matching)
        {
            if (matching.Match == 0 && matching.Left == true) return true; // Always match a null side
            if (matching.Match == Match && (!matching.Left) == Left) return true; else return false;
        }
    }
}
